// pair2tb version 10
// pair2tb.c -- convert a pairwise .maf file to a threaded blockset
// The given pairwise alignment must be single-cov in both sequences.
// the next two arguments are specie names, followed by a string or not: 
// "contigs+" or "contigs-", no mapping files are needed

#define VERSION 10

#include "util.h"
#include "maf.h"
#include "mz_scores.h"
#include "multi_util.h"
#include "seq.h"

void getMafBetween(struct mafFile* mf, char* filename) {
	int i, j, start, end, size, nali, flag1, flag2;
	SEQ * sf;
	char src[500], name[200], strand, comp_name[500], comp_src[500];
	unsigned char* s;
	struct mafAli **A, X, *a, *cp_list, *wk_list, *prev, *last, *b;
	struct mafComp C;


	X.next = NULL;
	X.score = 0.0;
	X.components = &C;
	C.next = NULL;
	C.strand = '+';
	
	sf = seq_open(filename);
	cp_list = mf->alignments;

	while (seq_read(sf)) {

	  s = SEQ_CHARS(sf);
	  if ( parseHeader(filename, sf, name, src, &start, &end, &strand, &size) != 0)
	    fatalf("Wrong format: %s", sf->header);
	  //end--;   // since the end itself is not included in the seq  <-- inclusive now

	  if ( cp_list == NULL) {
	    strcpy(comp_name, name);
	    if ( strcmp(name, src) != 0) {
	      strcat(comp_name, ".");
	      strcat(comp_name, src);
	    }
	    C.src = comp_name;
	    C.start = start - 1;
	    C.srcSize= size;
	    C.size = X.textSize = end-start+1;
	    C.text = copy_string(s);
	    mafWrite(stdout, &X);
	    free(C.text);
	    continue;
	  }

	  prev = a = cp_list;
	  wk_list = last = NULL;
	  while ( a != NULL) {
	    parseSrcName(a->components->src, comp_name, comp_src);
	    if ( strcmp(name, comp_name)==0 && strcmp(src, comp_src)==0 && a->components->start >= start-1 && a->components->start + a->components->size <= end) { // mv to wk_list
	      if ( a == cp_list) {
		cp_list = cp_list->next;
		a->next = NULL;
		b = a;
		a = prev = cp_list;
	      }
	      else {
		prev->next = a->next;
		a->next = NULL;
		b = a;
		a = prev->next;
	      }
	      if ( wk_list == NULL) 
		wk_list = last = b;
	      else {
		last->next = b;
		last = b;
	      }
	    }
	    else {
	      prev = a;
	      a = a->next;
	    }
	  }

	  A = mafArray(wk_list, &nali);

	  strcat(comp_name, ".");
	  strcat(comp_name, src);
	  C.src = comp_name;
	  C.srcSize = size;
	  
	  flag1 = start-1;
	  for (i=0; i<nali; i++) {
	    flag2 = A[i]->components->start - 1;
	    
	    if ( flag2 >= flag1 ) {
	      C.start = flag1;
	      X.textSize = C.size = flag2-flag1+1;
	      C.text = (char*)malloc((C.size+1)*sizeof(char));
  	      for (j=0; j<C.size; j++)
		C.text[j] = s[flag1-start+j+1];
	      C.text[C.size] = '\0';
	      mafWrite(stdout, &X);
	      free(C.text);
	    }
	    flag1 = A[i]->components->start + A[i]->components->size; 
	  }
	  if ( flag1 <= end-1 ) { // the last piece
	    C.start = flag1;
	    X.textSize = C.size = end-flag1;
	    C.text = (char*)malloc((C.size+1)*sizeof(char));
	    for (j=0; j<C.size; j++)
	      C.text[j] = s[flag1-start+j+1];
	    C.text[C.size] = '\0';
	    mafWrite(stdout, &X);
	    free(C.text);
	  }
	  for (i=0; i<nali; i++)
	    A[i] = NULL;
	  free(A);
	}
	seq_close(sf);
}

int main(int argc, char **argv) {
	struct mafFile *mf;
	struct mafAli *a;
	char cmd[500];

	sprintf(cmd, "pair2tb.v%d", VERSION);
	argv0 = cmd;
	if (argc != 4)
		fatal("-- convert a pairwise .maf file to a threaded blockset.\nargs: pairwise.maf seqfile1 seqfile2");

	mafWriteStart(stdout, cmd);

	mf = mafReadAll(argv[1], 1);

	for (a = mf->alignments ; a != NULL; a = a->next)
		mafWrite(stdout, a);

	getMafBetween(mf, argv[2]);
	mf = mafReadAll(argv[1], 0);
	flip_comps(mf->alignments);
	getMafBetween(mf, argv[3]);

	return 0;
}
