// tba version 10
// tba.c -- execute commands to run multiz and satellite programs, creating a
// threaded-block alignment

#define VERSION 10

// arg '-' means no execute, '+' means verbose

#include <unistd.h>
#include <sys/types.h>
#include "util.h"
#include "maf.h"
#define MZ "multiz"
#define MP "maf_project"
#define P2 "pair2tb"
#define GCD "get_covered"
//#define TBA_PREFIX "_tba_"
//#define DESTINATION "tba.maf"
#define DEFAULT_MIN_WIDTH "1"

#define SUFFIX ".maf"
#define NAMESIZE 500
#define BIG_BUF_SIZE 50000

char TBA_PREFIX[NAMESIZE], _tba_A[NAMESIZE], _tba_B[NAMESIZE], _tba_C[NAMESIZE], _tba_D[NAMESIZE], _tba_E[NAMESIZE], _tba_F[NAMESIZE], _tba_H[NAMESIZE], _tba_L[NAMESIZE], _tba_T[NAMESIZE], _tba_V[NAMESIZE], _tba_W[NAMESIZE], _tba_U[NAMESIZE], _tba_X[NAMESIZE], _tba_Y[NAMESIZE];

char DESTINATION[500];
char project_cmd[500];

typedef struct name_list_ {
	char *name;
	struct name_list_ *next;
} NameList, *NameListPtr;

static int execute, verbose;
static char mz[500];	// multiz command perhaps with R= and M= options

/* information at each node of a phylogenetic tree */
typedef struct tree_node_ {
	int id, type;
	NameListPtr names;	/* leaf-species names */
} TreeNode, *TreeNodePtr;

int get_rid_of_top(char* input_maf, char* output_maf) {
  struct mafFile* mf;
  struct mafAli* ali;
  struct mafComp* comp;
  FILE *fpw;

  if ( execute == 0)
    return 0;
  fpw = fopen(output_maf, "w");
  mf = mafOpen(input_maf, 0);
  while ( (ali=mafNext(mf))!=NULL) {
    comp=ali->components;
    if  (comp->next != NULL) {
      ali->components = comp->next;
      comp->next = NULL;
      mafCompFree(&comp);
      mafWrite(fpw, ali);
    }
    mafAliFree(&ali);
  }
  mafFileFree(&mf);
  fclose(fpw);
  return 0;
}

// do_cmd -- print and possible execute a command
void do_cmd(const char *fmt, ...) {
	char buf[BIG_BUF_SIZE];
	va_list ap;

        // I restructured this to support the x86_64 architecture
        // without using va_copy from C99. -rico

        va_start(ap, fmt);
        (void)vsprintf(buf, fmt, ap);
        va_end(ap);

        if (verbose) {
                (void)printf("%s\n", buf);
                (void)fflush(stdout);
        }
        if (execute) {
                if (system(buf) != 0)
                        fatalf("command '%s' failed", buf);
        }
}

#ifdef DEAD_CODE
// added 9/5/03
void iterate_porder(int id, char *name_list) {
	char cmd1[10000], cmd2[10000], cmd3[1000];
	int iterate = 1;

	sprintf(cmd1, "%s %stba%d %s > %stemp",
	  MD, TBA_PREFIX, id, name_list, TBA_PREFIX);
	sprintf(cmd2, "%s %stemp > %stba%d", MD, TBA_PREFIX, TBA_PREFIX, id);
	sprintf(cmd3, "mv %stemp %stba%d", TBA_PREFIX, TBA_PREFIX, id);

	while (iterate) {
		iterate = 0;
		if (verbose)
			printf("%s\n", cmd1);
		if (execute && system(cmd1) != 0)
			fatalf("command '%s' failed", cmd1);
		if (verbose)
			printf("%s\n", cmd2);
		if (execute && (iterate = system(cmd2)) != 0) {
			if (verbose)
				printf("%s\n", cmd3);
			system(cmd3);
		}
	}
}
#endif

static char *bz_cmd(char *x, char *y, int nbz, char **bz_file) {
	int i;
	char buf[500];

	sprintf(buf, "%s.%s%s", x, y, SUFFIX);
	for (i = 0; i < nbz; ++i)
		if (same_string(buf, bz_file[i]))
			return copy_string(buf);
	return NULL;
}

/* merge_trees - generate commands to merge the threaded-block alignments for
*  the two offspring of a node in the phylogenetic tree.
*/
static void merge_trees(TreeNodePtr x, TreeNodePtr y, int id, int nbz,
  char **bz_file) {
	NameListPtr n, n1, n2;
	int found_cmd, single_left, single_right, single1, single2;
	char *cmd, *mp = MP, *p2 = P2, *gcd = GCD, tmp_cmd[200];
	char _tba_head_F[NAMESIZE], _tba_head_E[NAMESIZE];

	if ((n1 = x->names) == NULL || (n2 = y->names) == NULL)
		fatal("merge_tree: empty tree");
	if (n1->next == NULL && n2->next == NULL) {
		if ((cmd = bz_cmd(n1->name, n2->name, nbz, bz_file)) == NULL)
			fatalf("no alignment found for %s and %s", n1->name,
			  n2->name);
		do_cmd("%s %s %s %s > %stba%d", p2, cmd, n1->name, n2->name,
		  TBA_PREFIX, id);
		return;
	}

	single_left = (n1->next == NULL);
	single_right = (n2->next == NULL);
	found_cmd = 0;
	single1=single2=0;
	for (n1 = x->names; n1 != NULL; n1 = n1->next)
	  for (n2 = y->names; n2 != NULL; n2 = n2->next) { 
	    if ((cmd = bz_cmd(n1->name, n2->name, nbz, bz_file))!= NULL) {
	      do_cmd("rm -f %s %s %s %s %s %s %s %s %s %s %s %s %s %s", _tba_Y, _tba_X, _tba_U, _tba_W, _tba_T, _tba_E, _tba_F, _tba_H, _tba_L, _tba_V, _tba_A, _tba_B, _tba_C, _tba_D);
	      do_cmd("touch %s %s %s %s %s %s %s %s %s %s %s %s %s %s", _tba_Y, _tba_X, _tba_U, _tba_W, _tba_T, _tba_E, _tba_F, _tba_H, _tba_L, _tba_V, _tba_A, _tba_B, _tba_C, _tba_D);
	      if ( !single_left) {
		sprintf(tmp_cmd, "%sleft.maf%d", TBA_PREFIX, id);
		do_cmd(project_cmd, mp, tmp_cmd, n1->name, _tba_A, _tba_B);
		if ( !single_right) {
		  do_cmd("%s %s %s 1 %s %s > %s", mz, _tba_B, cmd, _tba_Y, _tba_X, _tba_H);
		  do_cmd(project_cmd, mp, _tba_H, n2->name, _tba_U, _tba_B);
		  sprintf(tmp_cmd, "%sright.maf%d", TBA_PREFIX, id);
		  do_cmd(project_cmd, mp, tmp_cmd, n2->name, _tba_C, _tba_D);
		  do_cmd("%s %s %s 1 %s %s nohead >> %stba%d", mz, _tba_D, _tba_B, _tba_E, _tba_F, TBA_PREFIX, id);
		  do_cmd("cat %shead %s > %shead_F", TBA_PREFIX, _tba_F, TBA_PREFIX);
		  sprintf(_tba_head_F, "%shead_F", TBA_PREFIX);
		  do_cmd(project_cmd, mp, _tba_head_F, n2->name, _tba_W, _tba_H); 
		  get_rid_of_top(_tba_H, _tba_F);
		}
		else { // single_right
		  do_cmd("%s %s %s %s > %s", p2, cmd, n1->name, n2->name, _tba_D);
		  do_cmd(project_cmd, mp, _tba_D, n2->name, _tba_V, _tba_H);
		  if ( single2 == 0) {
		    do_cmd(project_cmd, mp, _tba_H, n1->name, _tba_C, _tba_D);
		    single2++;
		  }
		  else {
		    sprintf(tmp_cmd, "%sright.maf%d", TBA_PREFIX, id);
		    do_cmd(project_cmd, mp, tmp_cmd, n2->name, _tba_V, _tba_D);
		    do_cmd("%s %s %s > %s", gcd, _tba_H, _tba_D, _tba_L);
		    do_cmd(project_cmd, mp, _tba_L, n1->name, _tba_C, _tba_D);
		  }
                  do_cmd("%s %s %s 1 %s %s nohead >> %stba%d", mz, _tba_B, _tba_D, _tba_F, _tba_E, TBA_PREFIX, id);
		  do_cmd("cat %shead %s > %shead_E", TBA_PREFIX, _tba_E, TBA_PREFIX);
		  sprintf(_tba_head_E, "%shead_E", TBA_PREFIX);
		  do_cmd(project_cmd, mp, _tba_head_E, n1->name, _tba_T, _tba_H); 
		  get_rid_of_top(_tba_H, _tba_E);
		}
	      }
		
	      else if ( single_left) {
		do_cmd("%s %s %s %s > %s", p2, cmd, n1->name, n2->name, _tba_B);  
		do_cmd(project_cmd, mp, _tba_B, n1->name, _tba_V, _tba_H);
		if ( single1 == 0) {
		  do_cmd(project_cmd, mp, _tba_H, n2->name, _tba_A, _tba_B);
		  single1++;
		}
		else {
		  sprintf(tmp_cmd, "%sleft.maf%d", TBA_PREFIX, id);
		  do_cmd(project_cmd, mp, tmp_cmd, n1->name, _tba_V, _tba_B);
		  do_cmd("%s %s %s > %s", gcd, _tba_H, _tba_B, _tba_L);
		  do_cmd(project_cmd, mp, _tba_L, n2->name, _tba_A, _tba_B);
		}
		sprintf(tmp_cmd, "%sright.maf%d", TBA_PREFIX, id);
		do_cmd(project_cmd, mp, tmp_cmd, n2->name, _tba_C, _tba_D);
		do_cmd("%s %s %s 1 %s %s nohead >> %stba%d", mz, _tba_D, _tba_B, _tba_E, _tba_F, TBA_PREFIX, id);
		do_cmd("cat %shead %s > %shead_F", TBA_PREFIX, _tba_F, TBA_PREFIX);
		sprintf(_tba_head_F, "%shead_F", TBA_PREFIX);
		sprintf(tmp_cmd, "%stmp_F", TBA_PREFIX);
		do_cmd(project_cmd, mp, _tba_head_F, n2->name, _tba_W, tmp_cmd);
		get_rid_of_top(tmp_cmd, _tba_F);
	      }
	      do_cmd("cat %shead %s %s %s %s %s > %sleft.maf%d", TBA_PREFIX, _tba_A, _tba_Y, _tba_U, _tba_F, _tba_W, TBA_PREFIX, id);
	      do_cmd("cat %shead %s %s %s > %sright.maf%d", TBA_PREFIX, _tba_C, _tba_E, _tba_T, TBA_PREFIX, id);
	      found_cmd = 1;
	    }
	    if (!found_cmd) {
	      fprintf(stderr, "No alignments connect tree with leaves:\n");
	      for (n = x->names; n; n = n->next)
		fprintf(stderr, "  %s", n->name);
	      putc('\n', stderr);
	      fprintf(stderr, "and tree with leaves:\n");
	      for (n = y->names ; n; n = n->next)
		fprintf(stderr, "  %s", n->name);
	      putc('\n', stderr);
	      exit(1);
	    }
	  }
}

int main(int argc, char **argv) {
  char *q, buf[500], *p, name_list[10000], cmd[500], 
	  big_buf[BIG_BUF_SIZE], *bz_file[10000], maf_file[500];
	int i, j, top = -1, id = 0, nbz;
	TreeNode tree[1000];
	NameListPtr n;
	struct mafFile* maf;
	pid_t pid;

	sprintf(cmd, "tba.v%d", VERSION);
	argv0 = cmd;

	if (argc < 4)
	       fatalf("TBA -- threaded block alignment.\nargs: [+-] [R=?] [M=?] species-guid-tree maf-source destination");


	strcpy(DESTINATION, argv[argc-1]);

	pid = getpid();
	sprintf(TBA_PREFIX, "/tmp/_tba_%d_", pid);
	sprintf(_tba_A, "%sA", TBA_PREFIX);
	sprintf(_tba_B, "%sB", TBA_PREFIX);
	sprintf(_tba_C, "%sC", TBA_PREFIX);
	sprintf(_tba_D, "%sD", TBA_PREFIX);
	sprintf(_tba_E, "%sE", TBA_PREFIX);
	sprintf(_tba_F, "%sF", TBA_PREFIX);
	sprintf(_tba_H, "%sH", TBA_PREFIX);
	sprintf(_tba_L, "%sL", TBA_PREFIX);
	sprintf(_tba_T, "%sT", TBA_PREFIX);
	sprintf(_tba_W, "%sW", TBA_PREFIX);
	sprintf(_tba_U, "%sU", TBA_PREFIX);
	sprintf(_tba_V, "%sV", TBA_PREFIX);
	sprintf(_tba_X, "%sX", TBA_PREFIX);
	sprintf(_tba_Y, "%sY", TBA_PREFIX);

	sprintf(big_buf, "# %s", cmd);
	for (i = 1; i < argc; ++i) {
		if (strlen(big_buf) + strlen(argv[1]) + 2 >= BIG_BUF_SIZE)
			fatal("overflow = big_buf");
		strcat(big_buf, " ");
		strcat(big_buf, argv[i]);
	}

	strcpy(mz, MZ);
	strcat(mz, " ");
	verbose = 0;
	execute = 1;
	buf[0] = '\0';
	if (argc > 1 && same_string(argv[1], "-")) {
		strcat(buf, "-");
		execute = 0;
		verbose = 1;
		--argc;
		++argv;
	} else if (argc > 1 && same_string(argv[1], "+")) {
		strcat(buf, "+");
		verbose = 1;
		--argc;
		++argv;
	}

	while (argc > 1 && strchr("RM",argv[1][0]) && argv[1][1] == '=') {
		strcat(mz, argv[1]);
		strcat(mz, " ");
		--argc;
		++argv;
	}
	if (strchr(mz, 'M') == NULL) {
	  strcat(mz, "M=");
	  strcat(mz, DEFAULT_MIN_WIDTH);
	  strcat(mz, " ");
	}

	
	strcpy(project_cmd, "%s %s %s %s > %s");
	/*
	if ( argv[argc-1][0]=='P' && argv[argc-1][1]=='=') {
	  mapping_fn = argv[argc-1]+2;
	  strcat(project_cmd, " P=");
	  strcat(project_cmd, mapping_fn);
	  argc--;
	}
	else 
	  mapping_fn = NULL;
	*/


	if (argc == 5 && same_string(argv[2], "-f")) {
		char buf2[500], *s;
		FILE *fp = ckopen(argv[3], "r");

		for (nbz = 0; fgets(buf2, 500, fp) != NULL; ++nbz) {
			if ((s = strchr(buf2, '\n')) != NULL)
				*s = '\0';
			bz_file[nbz] = copy_string(buf2);
		}
		fclose(fp);
	} else {
		nbz = argc - 3;
		for (i = 0; i < nbz; ++i)
			bz_file[i] = argv[i+2];
	}
	

	do_cmd("rm -f %s", DESTINATION);
	do_cmd("echo \"##maf version=1 scoring=multiz\" > %shead",
	  TBA_PREFIX);

	do_cmd("echo \"##maf version=1 scoring=%s\" > %s", cmd, DESTINATION);
	do_cmd("echo \"%s\" >> %s", big_buf, DESTINATION);
	//	do_cmd("%s version >> %s", MZ, DESTINATION);

	for (q = argv[1]; *q != '\0'; ++q) {
		if (*q == '(') {
			if (++top >= 1000)
				fatal("parse_tree: stack overflow");
			tree[top].type = '(';
		} else if (*q == ')') {
			if (top < 1 || tree[top].type != 0 ||
				       tree[top-1].type != '(') {
				q[1] = '\0';
				fatalf("parse error: %s", argv[1]);
			}
			tree[top-1] = tree[top];
			--top;
		} else if (isalpha(*q)) {
			/* leaf species */
			if (++top >= 1000)
				fatal("parse_tree: stack overflow");
			p = buf;
			while (isalpha(*q) || isdigit(*q) || *q == '_' ||
			  *q == '.')
				*p++ = *q++;
			--q;
			*p = '\0';
			n = ckalloc(sizeof(NameList));
			n->name = copy_string(buf);
			n->next = NULL;
			tree[top].id = -1;
			tree[top].type = 0;
			tree[top].names = n;
		} else if (*q != ' ')
			fatalf("improper character in tree specification: %c",
				*q);
		if (top > 0 && tree[top-1].type == 0 && tree[top].type == 0) {
			// Process an internal node by merging its children.
			// One optimization is to identify nodes where both
			// children are leaves, to avoid a use of multiz.
			i = tree[top-1].id;
			j = tree[top].id;
			if (i >= 0)
			  do_cmd("mv %stba%d %sleft.maf%d",
				  TBA_PREFIX, i, TBA_PREFIX, id);
			if (j >= 0)
			  do_cmd("mv %stba%d %sright.maf%d",
				  TBA_PREFIX, j, TBA_PREFIX, id);
			do_cmd("cp %shead %stba%d",
			  TBA_PREFIX, TBA_PREFIX, id);
			merge_trees(&(tree[top-1]), &(tree[top]), id, nbz,
			  bz_file);
			if (i >= 0 || j >= 0) {
			  sprintf(maf_file, "%sleft.maf%d", TBA_PREFIX, id);
			  if ( execute == 1) {
			    
			    maf = mafOpen(maf_file, 0);
			    if ( mafNext(maf)!= NULL) {
			      mafFileFree(&maf);
			      do_cmd("grep -v maf %sleft.maf%d >> %stba%d", TBA_PREFIX, id, TBA_PREFIX, id);
			    }
			    else
			      mafFileFree(&maf);
			    sprintf(maf_file, "%sright.maf%d", TBA_PREFIX, id);
			    maf = mafOpen(maf_file, 0);
			    if ( mafNext(maf)!=NULL) {
			      mafFileFree(&maf);
			      do_cmd("grep -v maf %sright.maf%d >> %stba%d", TBA_PREFIX, id, TBA_PREFIX, id);
			    }
			    else
			      mafFileFree(&maf);
			  }
			  name_list[0] = '\0';
			  for (n = tree[top-1].names; n != NULL;
			       n = n->next) {
			    strcat(name_list, " ");
			    strcat(name_list, n->name);
			  }
			  strcat(name_list, " +");
			  for (n = tree[top].names; n != NULL;
				  n = n->next) {
			    strcat(name_list, " ");
			    strcat(name_list, n->name);
			  }
			}
			/* merge the lists of names */
			n = tree[top-1].names;
			if (n == NULL)
				fatal("empty list of names");
			while (n->next != NULL)
				n = n->next;
			n->next = tree[top].names;
			--top;
			tree[top].id = id++;
		}
	}
	//printf("\n");
	if ((p = strchr(name_list, '+')) != NULL)
		*p = ' ';
	//	do_cmd("%s %stba%d %s nohead >> %s", mo, TBA_PREFIX, id-1, name_list,
	//	  DESTINATION);
		do_cmd("mv %stba%d %s", TBA_PREFIX, id-1, DESTINATION);
		do_cmd("rm %s*", TBA_PREFIX);
	if (top > 0)
		fatal("tree specification contains too many '('");
	if (top != 0 || tree[0].type != 0)
		fatal("tree specification is improper");
	return 0;
}
